from OFS.PropertyManager import PropertyManager
from math import *

class GraphObject(PropertyManager):
    '''Abstract class defining the property of any object that can be drawn
       inside the ZQuest Graph.'''

    _properties = ({'id': 'frame', 'type': 'object', 'default': None}, )
    _propOrder = ['frame']


    def __init__(self, *props, **kw):
        '''Very general GraphObject Constructor; return None'''
        self._initProperties(props, kw)
        

    def _initProperties(self, props, kw):
        '''Initialize the properties.'''
        
        for id in self.propertyIds():
            if id in kw.keys():
                self._updateProperty(id, kw[id])
            else:
                self._updateProperty(id, self.getDefaultValue(id))

        for id in self._propOrder:
            index = self._propOrder.index(id)
            if index < len(props):
                self._updateProperty(id, props[index])


    def getDefaultValue(self, id):
        '''Get the default value of the property with id=id; return object'''
        for prop in self._properties:
            if prop['id'] == id:
                return prop['default']

        return None


    def getPropertyDict(self):
        ''' '''
        dict = {}
        for prop in self._properties:
            dict[prop['id']] = getattr(self, prop['id'])

        return dict
        

    def printProperties(self):
        '''Print the properties to stdin'''
        print self.propertyItems()



class Arrow(GraphObject):
    '''The Arrow allows to draw an arrow on the frame. There are several types '''

    _properties = ( GraphObject._properties +
                    ({'id': 'start',     'type': 'tuple', 'default': (0, 0)},
                     {'id': 'end',       'type': 'tuple', 'default': (0, 0)},
                     {'id': 'size',      'type': 'int',   'default': 5},
                     {'id': 'type',      'type': 'int',   'default': 0},
                     {'id': 'headAngle', 'type': 'int',   'default': pi/3},
                     {'id': 'color',     'type': 'tuple', 'default': (0, 0, 0)},
                     {'id': 'fillColor', 'type': 'tuple', 'default': (0, 0, 0)}) )

    _propOrder = ['frame', 'start', 'end']

    types = ('simple', 'filled')


    def draw(self):
        '''Draw the arrow on the frame; return None'''
        self.frame.line(self.start+self.end, fill=self.color)

        # determine the angle of the line just drawn
        try:
            theta = atan(float(self.end[0]-self.start[0])/float(self.end[1]-self.start[1]))
        except ZeroDivisionError:
            theta = 0

        # determine to which quadrant the arrow points and adjust the angle 
        deltaX = self.end[0]-self.start[0]
        deltaY = self.end[1]-self.start[1]

        if (deltaX < 0 and deltaY < 0):
            theta = theta + pi/2
        elif (deltaX < 0 and deltaY > 0):
            theta = theta - pi/2
        elif (deltaX >= 0 and deltaY < 0):
            theta = theta + pi/2
        elif (deltaX >= 0 and deltaY > 0):
            theta = theta - pi/2
        elif (deltaX < 0 and deltaY == 0):
            theta = theta - pi

        # determine the coordinates tp draw the arrow head lines
        theta1 = theta  + pi
        theta1 = theta1 + self.headAngle/2
        head1 = (self.end[0]+int(cos(theta1)*self.size),
                 self.end[1]-int(sin(theta1)*self.size))

        theta2 = theta  - pi
        theta2 = theta2 - self.headAngle/2
        head2 = (self.end[0]+int(cos(theta2)*self.size),
                 self.end[1]-int(sin(theta2)*self.size))

        # based on the selected arrow type, draw the arrow head
        if self.type == 0:
            self.frame.line(self.end + head1, fill=self.color)
            self.frame.line(self.end + head2, fill=self.color)
        elif self.type == 1:
            self.frame.polygon(self.end + head1 + head2, outline=self.color,
                               fill=self.fillColor)


class Vector(GraphObject):
    '''Draw a vector into the coordinate system'''

    _properties = ( ({'id': 'coord',     'type': 'object', 'default': None},
                     {'id': 'start',     'type': 'tuple',  'default': (0, 0)},
                     {'id': 'end',       'type': 'tuple',  'default': (0, 0)},
                     {'id': 'size',      'type': 'int',    'default': 5},
                     {'id': 'type',      'type': 'int',    'default': 0},
                     {'id': 'headAngle', 'type': 'int',    'default': pi/3},
                     {'id': 'color',     'type': 'tuple',  'default': (0, 0, 0)},
                     {'id': 'fillColor', 'type': 'tuple',  'default': (0, 0, 0)}) )

    _propOrder = ['coord', 'start', 'end']

    def draw(self):
        '''Draw the vector; return None'''
        dict = self.getPropertyDict()
        dict['frame'] = self.coord.frame

        # convert the coordinates to pixels
        dict['start'] = self.coord.getPixelCoordinates(self.start)
        dict['end']   = self.coord.getPixelCoordinates(self.end)

        # well, let's just reuse the arrow to draw our vector
        apply(Arrow, (), dict).draw()



class Axis(GraphObject):
    '''Draw any axis into the frame. (This is an abstract class.)'''

    _properties = ( GraphObject._properties +
                    ({'id': 'start',       'type': 'float',    'default': -10},
                     {'id': 'end',         'type': 'float',    'default':  10},
                     {'id': 'stepSize',    'type': 'float',    'default':   2},
                     {'id': 'offset',      'type': 'tuple',    'default': (100, 100)},
                     {'id': 'label',       'type': 'string',   'default':  ''},
                     {'id': 'scale',       'type': 'int',      'default':  10},
                     {'id': 'labelFormat', 'type': 'string',   'default':  '%.2f'},
                     {'id': 'labelColor',  'type': 'tuple',    'default': (0, 0, 255)},
                     {'id': 'valueColor',  'type': 'tuple',    'default': (0, 255, 0)},
                     {'id': 'color',       'type': 'tuple',    'default': (0, 0, 0)}) )

    _propOrder = ['frame', 'start', 'end', 'stepSize', 'offset']

    def draw(self):
        pass



class XAxis(Axis):
    '''Implementation of the x-axis'''

    def draw(self):
        '''Draw the axis on the frame'''

        # Let's find the length of the axis in pixel
        endPixel   = self.offset[0] + (self.end-self.start)*self.scale + 2*5
        
        # First draw the line that represents the axis
        self.frame.line(self.offset+(endPixel, self.offset[1]), fill=self.color)

        # Draw the markers
        marker = self.start
        while marker <= self.end:
            markerPos = self.offset[0] + 5 + (marker-self.start)*self.scale

            # Draw the marker line
            self.frame.line((markerPos, self.offset[1]-3,
                             markerPos, self.offset[1]+4), fill=self.color)

            # Draw the marker value
            valueString = self.labelFormat %marker
            valueSize   = self.frame.textsize(valueString)
            self.frame.text((markerPos-valueSize[0]/2, self.offset[1]+3), valueString,
                            fill=self.valueColor)

            marker = marker + self.stepSize
    
        # Draw the label
        labelSize   = self.frame.textsize(self.label)
        self.frame.text((endPixel-labelSize[0], self.offset[1]+labelSize[1]+3), self.label,
                            fill=self.labelColor)



class YAxis(Axis):
    '''Implementation of the y-axis'''

    def draw(self):
        '''Draw the axis on the frame'''

        # Let's find the length of the axis in pixel
        endPixel   = self.offset[1] - (self.end-self.start)*self.scale - 2*5
        
        # First draw the line that represents the axis
        self.frame.line(self.offset+(self.offset[0], endPixel))


        # Draw the markers
        marker = self.start
        while marker <= self.end:
            markerPos = self.offset[1] - 5 - (marker-self.start)*self.scale

            # Draw the marker line
            self.frame.line((self.offset[0]-3, markerPos, 
                             self.offset[0]+3, markerPos), fill=self.color)

            # Draw the marker value
            valueString = self.labelFormat %marker
            valueSize   = self.frame.textsize(valueString)
            self.frame.text((self.offset[0]-4-valueSize[0], markerPos-valueSize[1]/2,),
                            valueString, fill=self.valueColor)

            marker = marker + self.stepSize
    
        # Draw the label
        labelSize   = self.frame.textsize(self.label)
        labelX = self.offset[0]-labelSize[0]*2/3
        if labelX < 0:  # quick fix, since this would run out of the frame regulary
            labelX = 0
        self.frame.text((labelX, endPixel-labelSize[1]), self.label,
                        fill=self.labelColor)



class Grid(GraphObject):
    '''Draw a grid on the .'''

    _properties = ( GraphObject._properties +
                    ({'id': 'start',      'type': 'tuple',    'default': -10},
                     {'id': 'end',        'type': 'tuple',    'default':  10},
                     {'id': 'stepSize',   'type': 'tuple',    'default':  (2, 2)},
                     {'id': 'offset',     'type': 'tuple',    'default': (100, 100)},
                     {'id': 'scale',      'type': 'int',      'default':  10},
                     {'id': 'color',      'type': 'tuple',    'default': (0, 0, 0)}) )

    _propOrder = ['frame', 'start', 'end', 'stepSize', 'offset']


    def drawX(self):
        '''Draw the y component'''

        # Let's find the length of the axis in pixel
        endYPixel   = self.offset[1] - (self.end[1]-self.start[1])*self.scale[1] - 5*2
        
        # Draw the x-axis grid
        marker = self.start[0]
        while marker <= self.end[0]:
            markerPos = self.offset[0] + (marker-self.start[0])*self.scale[0] + 5

            # Draw a grid line
            self.frame.line((markerPos, self.offset[1], markerPos, endYPixel), fill=self.color)

            marker = marker + self.stepSize[0]


    def drawY(self):
        '''Draw the x component'''

        # Let's find the length of the axis in pixel
        endXPixel   = self.offset[0] + (self.end[0]-self.start[0])*self.scale[0] + 5*2
        
        # Draw the y-axis grid
        marker = self.start[1]
        while marker <= self.end[1]:
            markerPos = self.offset[1] - (marker-self.start[1])*self.scale[1] - 5

            # Draw a grid line
            self.frame.line((self.offset[0], markerPos, endXPixel, markerPos), fill=self.color)

            marker = marker + self.stepSize[1]


    def draw(self):
        '''Draw the x and y components of the grid.'''
        self.drawX()
        self.drawY()


class CoordinateSystem(GraphObject):
    '''The Coordinate System describes how it should present itself inside the frame.'''

    _properties = ( GraphObject._properties +
                    ({'id': 'start',      'type': 'tuple',    'default':      -10},
                     {'id': 'end',        'type': 'tuple',    'default':       10},
                     {'id': 'stepSize',   'type': 'tuple',    'default': ( 5,  5)},
                     {'id': 'scale',      'type': 'tuple',    'default': (10, 10)},
                     {'id': 'intersect',  'type': 'tuple',    'default': ( 0,  0)},
                     {'id': 'offset',     'type': 'tuple',    'default': ( 0,  0)},  # value in pixels
                     {'id': 'title',      'type': 'string',   'default':       ''},
                     {'id': 'labels',     'type': 'tuple',    'default': ('', '')},
                     {'id': 'hasGrid',    'type': 'int',      'default':        0},
                     {'id': 'hasBorder',  'type': 'int',      'default':        0},
                     {'id': 'gridColor',  'type': 'tuple',    'default': (220, 220, 220)},
                     {'id': 'color',      'type': 'tuple',    'default': (  0,   0,   0)}) )

    _propOrder = ['frame', 'start', 'end', 'stepSize']


    def __init__(self, *props, **kw):
        '''Constructor of the Coordinate System'''

        self._initProperties(props, kw)

        xOffset = (self.offset[0]+15, self.offset[1]+20+(self.end[1]-self.intersect[1])*self.scale[1])
        yOffset = (self.offset[0]+20+(self.intersect[0]-self.start[0])*self.scale[0],
                   self.offset[1]+25+(self.end[1]-self.start[1])*self.scale[1])
                   
        self.xaxis = XAxis(self.frame, self.start[0], self.end[0], self.stepSize[0], xOffset, 
                           label=self.labels[0], scale=self.scale[0])
        self.yaxis = YAxis(self.frame, self.start[1], self.end[1], self.stepSize[1], yOffset,
                           label=self.labels[1], scale=self.scale[1])
        self.grid = Grid(self.frame, self.start, self.end, self.stepSize, (xOffset[0], yOffset[1]),
                         scale=self.scale, color=self.gridColor)
        

    def getPixelCoordinates(self, point):
        '''Get the pixel value of the point'''
        return (self.offset[0] + 20 + (point[0] - self.start[0])*self.scale[0],
                self.offset[1] + 20 + (self.end[1] - point[1])*self.scale[1])

    
    def _drawTitle(self):
        '''Draw the title of the graph'''
        position = (self.offset[0]+15,
                    self.offset[1]+2)
        self.frame.text(position, self.title)
        

    def _drawBorder(self):
        '''Draw a border around the coordinate system'''
        self.frame.rectangle(self.offset + (self.offset[0]+40+(self.end[0]-self.start[0])*self.scale[0],
                                            self.offset[1]+40+(self.end[1]-self.start[1])*self.scale[1]),
                             outline=self.color)

        
    def draw(self):
        '''Draw the Coordinate System'''
        self.frame.setfont(self.frame.fonts['Arial 8'])
        if self.hasGrid:
            self.grid.draw()
        self.xaxis.draw()
        self.yaxis.draw()

        if self.title:
            self._drawTitle()

        if self.hasBorder:
            self._drawBorder()


        
class DataPoint(GraphObject):
    '''This class represents a single data point in a coordinate system'''

    _properties = ( ({'id': 'coord',       'type': 'object', 'default':  None},
                     {'id': 'point',       'type': 'tuple',  'default':  (0, 0)},
                     {'id': 'type',        'type': 'int',    'default':  1},
                     {'id': 'size',        'type': 'int',    'default':  3},
                     {'id': 'label',       'type': 'string', 'default':  ''},
                     {'id': 'showCoord',   'type': 'int',    'default':  0},
                     {'id': 'coordFormat', 'type': 'string', 'default':  '(%.2f, %.2f)'},
                     {'id': 'color',       'type': 'tuple',  'default':  (0, 0, 0)},
                     {'id': 'labelColor',  'type': 'tuple',  'default': (0, 0, 0)}) )

    _propOrder = ['coord', 'point']

    types = ('dot', 'cross', 'plus', 'rectangle')


    def getPixelCoordinates(self):
        '''Get the pixel-based coordinates of the point'''
        return self.coord.getPixelCoordinates(self.point)


    def draw(self):
        '''Draw the data point in the coordinate system'''
        # get the pixel values for the data points location
        xpos, ypos = self.getPixelCoordinates()

        # Depending on the type of the point, draw it
        if self.type == 0:
            self.coord.frame.point((xpos, ypos), fill=self.color)
        elif self.type == 1:
            self.coord.frame.line((xpos-self.size+1, ypos-self.size+1,
                                   xpos+self.size,   ypos+self.size), fill=self.color)
            self.coord.frame.line((xpos-self.size+1, ypos+self.size-1,
                                   xpos+self.size,   ypos-self.size), fill=self.color)
        elif self.type == 2:
            self.coord.frame.line((xpos-self.size-1, ypos,
                                   xpos+self.size-1, ypos), fill=self.color)
            self.coord.frame.line((xpos, ypos-self.size-1,
                                   xpos, ypos+self.size), fill=self.color)
        elif self.type == 3:
            self.coord.frame.rectangle((xpos-self.size-1, ypos-self.size-1,
                                        xpos+self.size,   ypos+self.size), outline=self.color)

        if self.label or self.showCoord:
            label = self.label
            if self.showCoord:
                label = label + self.coordFormat %self.point
            self.coord.frame.text((xpos+self.size, ypos+self.size/2), label, fill=self.labelColor)



class FunctionCurve(GraphObject):
    '''This class will draw a curve of a specified function into the coordinate system'''

    _properties = ( ({'id': 'coord',          'type': 'object', 'default':  None},
                     {'id': 'function',       'type': 'string', 'default':  'sin(x)'},
                     {'id': 'range',          'type': 'int',    'default':  (-10, 10)},
                     {'id': 'stepSize',       'type': 'int',    'default':  .2},
                     {'id': 'dataPointType',  'type': 'int',    'default':  0},
                     {'id': 'isConnected',    'type': 'int',    'default':  1},
                     {'id': 'animate',        'type': 'int',    'default':  0},
                     {'id': 'label',          'type': 'string', 'default':  ''},
                     {'id': 'lineColor',      'type': 'tuple',  'default':  (255, 0, 0)},
                     {'id': 'dataPointColor', 'type': 'tuple',  'default':  (0, 0, 0)},
                     {'id': 'labelColor',     'type': 'tuple', 'default':   (0, 0, 0)}) )

    _propOrder = ['coord', 'function', 'range']


    def draw(self):
        '''Draw the function in the coordinate system'''

        x = self.range[0]
        previous = None
        
        while x <= self.range[1]:
            point = DataPoint(self.coord, (x, eval(self.function)),
                              type=self.dataPointType, color=self.dataPointColor)
            point.draw()
            if previous and self.isConnected:
                self.coord.frame.line(previous.getPixelCoordinates()+point.getPixelCoordinates(),
                                      self.lineColor)

            if self.animate:
                self.coord.frame.animation.addFrame()
                
            previous = point
            x = x + self.stepSize
            


class RandomDataCurve(GraphObject):
    '''This Class plots a set of given points'''

    _properties = ( ({'id': 'coord',          'type': 'object', 'default':  None},
                     {'id': 'data',           'type': 'tuple',  'default':  ()},
                     {'id': 'dataPointType',  'type': 'int',    'default':  0},
                     {'id': 'isConnected',    'type': 'int',    'default':  1},
                     {'id': 'animate',        'type': 'int',    'default':  0},
                     {'id': 'label',          'type': 'string', 'default':  ''},
                     {'id': 'lineColor',      'type': 'tuple',  'default':  (255, 0, 0)},
                     {'id': 'dataPointColor', 'type': 'tuple',  'default':  (0, 0, 0)},
                     {'id': 'labelColor',     'type': 'tuple',  'default':  (0, 0, 0)}) )

    _propOrder = ['coord', 'data']


    def draw(self):
        '''Draws the curve on the coordinate system'''
        previous = None
        for point in self.data:
            point.type  = self.dataPointType
            point.color = self.dataPointColor
            point.draw()
            if previous and self.isConnected:
                self.coord.frame.line(previous.getPixelCoordinates()+point.getPixelCoordinates(),
                                      self.lineColor)

            if self.animate:
                addFrame()

            previous = point
